'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var borel = require('borel');
var veho = require('veho');
var farbe = require('farbe');
var hatsu = require('hatsu');
var typen = require('typen');
var palett = require('palett');

/**
 * Create a chalk from a hsl array
 * @param {[number,number,number]} hsl
 * @returns {function}
 */

const tube = hsl => hatsu.Hatsu.hsl(hsl);

const {
  isNumeric: isNum
} = typen.Num;

const amp = (x, m, rto, min) => (x - m) * rto + min;
const makeHatsu = (x, m, [rH, rS, rL], [mH, mS, mL]) => {
  var _ref;

  return _ref = [amp(x, m, rH, mH), amp(x, m, rS, mS), amp(x, m, rL, mL)], tube(_ref);
};

const {
  mutateMap,
  map: mapAr
} = veho.Ar;

/**
 *
 * @param {*[]} ar
 * @param {number|*} m
 * @param {number|*} d
 * @param {[number,number,number]} min
 * @param {[number,number,number]} dif
 * @param {function} naChalk
 * @param {boolean} retFn
 * @param {boolean} mutate
 * @returns {function[]|*[]}
 */

const vecVisual = (ar, {
  val: {
    min: m,
    dif: d
  },
  pal: {
    min,
    dif
  },
  naChalk,
  retFn,
  mutate
}) => {
  const map = mutate ? mutateMap : mapAr;

  if (isNaN(d) || !d || dif.every(x => !x)) {
    var _min;

    const soleChalk = (_min = min, tube(_min));
    return retFn ? map(ar, x => isNum(x) ? soleChalk : naChalk) : map(ar, x => {
      var _x, _x2;

      return isNum(x) ? (_x = x, soleChalk(_x)) : (_x2 = x, naChalk(_x2));
    });
  }

  const rto = mapAr(dif, _ => _ / d, 3);
  return retFn ? map(ar, x => isNum(x) ? makeHatsu(x, m, rto, min) : naChalk) : map(ar, x => {
    var _x3, _x4;

    return isNum(x) ? (_x3 = x, makeHatsu(x, m, rto, min)(_x3)) : (_x4 = x, naChalk(_x4));
  });
};

/**
 *
 * @param {*[][]} mx
 * @param {number|*} m
 * @param {number|*} d
 * @param {[number,number,number]} min
 * @param {[number,number,number]} dif
 * @param {function} naChalk
 * @param {boolean} retFn
 * @param {boolean} mutate
 * @returns {function[]|*[]}
 */

const mxVisual = (mx, {
  val: {
    min: m,
    dif: d
  },
  pal: {
    min,
    dif
  },
  naChalk,
  retFn,
  mutate
}) => {
  const map = mutate ? mutateMap : mapAr;

  if (isNaN(d) || !d || dif.every(el => !el)) {
    var _min;

    const soleChalk = (_min = min, tube(_min));
    return retFn ? map(mx, el => isNum(el) ? soleChalk : naChalk) : map(mx, el => {
      var _el, _el2;

      return isNum(el) ? (_el = el, soleChalk(_el)) : (_el2 = el, naChalk(_el2));
    });
  }

  const rto = mapAr(dif, _ => _ / d, 3);
  return retFn ? map(mx, el => isNum(el) ? makeHatsu(el, m, rto, min) : naChalk) : map(mx, el => {
    var _el3, _el4;

    return isNum(el) ? (_el3 = el, makeHatsu(el, m, rto, min)(_el3)) : (_el4 = el, naChalk(_el4));
  });
};
/**
 *
 * @param {*[][]} mx
 * @param {number} y
 * @param {number|*} m
 * @param {number|*} d
 * @param {[number,number,number]} min
 * @param {[number,number,number]} dif
 * @param {function} naChalk
 * @param {boolean} retFn
 * @param {boolean} mutate
 * @returns {function[]|*[]}
 */

const mxVisualCol = (mx, y, {
  val: {
    min: m,
    dif: d
  },
  pal: {
    min,
    dif
  },
  naChalk,
  retFn,
  mutate
}) => {
  const mapCol = mutate ? veho.Mx.mutateCol : veho.Mx.mapCol;

  if (isNaN(d) || !d || dif.every(x => !x)) {
    var _min2;

    const soleChalk = (_min2 = min, tube(_min2));
    return retFn ? mapCol(mx, y, el => isNum(el) ? soleChalk : naChalk) : mapCol(mx, y, el => {
      var _el5, _el6;

      return isNum(el) ? (_el5 = el, soleChalk(_el5)) : (_el6 = el, naChalk(_el6));
    });
  }

  const rto = mapAr(dif, _ => _ / d, 3);
  return retFn ? mapCol(mx, y, el => isNum(el) ? makeHatsu(el, m, rto, min) : naChalk) : mapCol(mx, y, el => {
    var _el7, _el8;

    return isNum(el) ? (_el7 = el, makeHatsu(el, m, rto, min)(_el7)) : (_el8 = el, naChalk(_el8));
  });
};

const marks = {
  fresh: {
    max: palett.Palett.lightGreen.accent_3,
    min: palett.Palett.red.accent_2,
    na: palett.Greys.brown.base
  },
  subtle: {
    max: palett.Greys.grey.lighten_5,
    min: palett.Greys.grey.darken_1,
    na: palett.Palett.indigo.lighten_2
  }
};

const {
  fromHex,
  toHsl
} = farbe.Rgb;

const parseHsl = color => {
  var _ref, _color;

  return typeof color === 'string' ? (_ref = (_color = color, fromHex(_color)), toHsl(_ref)) : color;
};
/**
 *
 * @param max
 * @param min
 * @returns {{dif: [number,number,number], min: [number,number,number]}}
 */


const colorBound = (max, min) => ({
  min,
  dif: veho.Ar.mutateMap(max, (a, i) => a - min[i], 3)
});

class Visual {
  /**
   *
   * @param {*[]} arr
   * @param {string|number[]} [max]
   * @param {string|number[]} [min]
   * @param {string|number[]} [na]
   * @param {boolean} [mutate=true]
   * @param {boolean} [retFn=false]
   */
  static vector(arr, {
    mark: {
      max = marks.fresh.max,
      min = marks.fresh.min,
      na = marks.fresh.na
    } = {},
    mutate = false,
    retFn = false
  } = {}) {
    var _max, _min, _ref2, _na;

    return vecVisual(arr, {
      val: borel.Stat.bound(arr, {
        dif: true,
        level: 2
      }),
      pal: colorBound((_max = max, parseHsl(_max)), (_min = min, parseHsl(_min))),
      naChalk: (_ref2 = (_na = na, parseHsl(_na)), tube(_ref2)),
      mutate,
      retFn
    });
  }
  /**
   *
   * @param {*[][]} mx
   * @param {string|number[]} [max]
   * @param {string|number[]} [min]
   * @param {string|number[]} [na]
   * @param {number} [direct=1] - 0:p-wise, 1:r-wise, 2:c-wise
   * @param {boolean} [mutate=true]
   * @param {boolean} [retFn=false]
   * @returns {*}
   */


  static matrix(mx, {
    mark: {
      max = marks.subtle.max,
      min = marks.subtle.min,
      na = marks.subtle.na
    } = {},
    direct = 1,
    mutate = false,
    retFn = false
  } = {}) {
    var _max2, _min2, _ref3, _na2;

    if (!veho.Mx.isMat(mx)) throw new Error('Not valid matrix');

    switch (direct) {
      case 2:
        const mapCols = veho.Mx.mapColumns;
        return mapCols(mx, c => Visual.vector(c, {
          mark: {
            max,
            min,
            na
          },
          mutate,
          retFn
        }));

      case 1:
        const mapRows = mutate ? mutateMap : mapAr;
        return mapRows(mx, r => Visual.vector(r, {
          mark: {
            max,
            min,
            na
          },
          mutate,
          retFn
        }));

      case 0:
      default:
        return mxVisual(mx, {
          val: borel.StatMx.bound(mx, {
            dif: true,
            level: 2
          }),
          pal: colorBound((_max2 = max, parseHsl(_max2)), (_min2 = min, parseHsl(_min2))),
          naChalk: (_ref3 = (_na2 = na, parseHsl(_na2)), tube(_ref3)),
          mutate,
          retFn
        });
    }
  }
  /**
   *
   * @param {*[][]} mx
   * @param {number} y
   * @param {string|number[]} [max]
   * @param {string|number[]} [min]
   * @param {string|number[]} [na]
   * @param {boolean} deep
   */


  static column(mx, y, {
    mark: {
      max = marks.fresh.max,
      min = marks.fresh.min,
      na = marks.fresh.na
    } = {},
    mutate = false,
    retFn = false
  } = {}) {
    var _max3, _min3, _ref4, _na3;

    return mxVisualCol(mx, y, {
      val: borel.StatMx.boundCol(mx, y, {
        dif: true,
        level: 2
      }),
      pal: colorBound((_max3 = max, parseHsl(_max3)), (_min3 = min, parseHsl(_min3))),
      naChalk: (_ref4 = (_na3 = na, parseHsl(_na3)), tube(_ref4)),
      mutate,
      retFn
    });
  }

}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

var _Object$entries$map;
class Pal {}

_defineProperty(Pal, "bases", (_Object$entries$map = Object.entries(palett.Palett).map(([name, {
  base
}]) => [name, base]), veho.Ob.fromEntries(_Object$entries$map)));

_defineProperty(Pal, "marks", marks);

var ansiRegex = options => {
  options = Object.assign({
    onlyFirst: false
  }, options);
  const pattern = ['[\\u001B\\u009B][[\\]()#;?]*(?:(?:(?:[a-zA-Z\\d]*(?:;[-a-zA-Z\\d\\/#&.:=?%@~_]*)*)?\\u0007)', '(?:(?:\\d{1,4}(?:;\\d{0,4})*)?[\\dA-PR-TZcf-ntqry=><~]))'].join('|');
  return new RegExp(pattern, options.onlyFirst ? undefined : 'g');
};

const regex = new RegExp(ansiRegex().source);

var hasAnsi = string => regex.test(string);

const stripAnsi = string => typeof string === 'string' ? string.replace(ansiRegex(), '') : string;

var stripAnsi_1 = stripAnsi;
var default_1 = stripAnsi;
stripAnsi_1.default = default_1;

const regex$1 = '[\uD800-\uDBFF][\uDC00-\uDFFF]';

var astralRegex = opts => opts && opts.exact ? new RegExp(`^${regex$1}$`) : new RegExp(regex$1, 'g');

const stringLength = string => stripAnsi_1(string).replace(astralRegex(), ' ').length;

var stringLength_1 = stringLength; // TODO: Remove this for the next major release

var default_1$1 = stringLength;
stringLength_1.default = default_1$1;

const Presets = {
  fresh: {
    max: palett.Palett.lightGreen.accent_3,
    min: palett.Palett.red.accent_2,
    na: palett.Greys.brown.base
  },
  subtle: {
    max: palett.Greys.grey.lighten_5,
    min: palett.Greys.grey.darken_1,
    na: palett.Palett.indigo.lighten_2
  }
};

/**
 * Create a hatsu tube from a hsl array
 * @param {[number,number,number]} hsl
 * @returns {function}
 */

const tube$1 = hsl => hatsu.Hatsu.hsl(hsl);

const {
  isNumeric: isNum$1
} = typen.Num;

const amp$1 = (x, m, rto, min) => (x - m) * rto + min;

const toHatsu = (x, m, [rH, rS, rL], [mH, mS, mL]) => {
  var _ref;

  return _ref = [amp$1(x, m, rH, mH), amp$1(x, m, rS, mS), amp$1(x, m, rL, mL)], tube$1(_ref);
};

const {
  mutateMap: mutateMap$1,
  map: mapAr$1
} = veho.Ar;
/**
 *
 * @param {*[]} ar
 * @param {number|*} m
 * @param {number|*} d
 * @param {[number,number,number]} min
 * @param {[number,number,number]} dif
 * @param {function} naChalk
 * @param {boolean} retFn
 * @param {boolean} mutate
 * @returns {function[]|*[]}
 */

const vecVisual$1 = (ar, {
  val: {
    min: m,
    dif: d
  },
  pal: {
    min,
    dif
  },
  naChalk,
  retFn,
  mutate
}) => {
  const map = mutate ? mutateMap$1 : mapAr$1;

  if (isNaN(d) || !d || dif.every(x => !x)) {
    var _min;

    const soleChalk = (_min = min, tube$1(_min));
    return retFn ? map(ar, x => isNum$1(x) ? soleChalk : naChalk) : map(ar, x => {
      var _x, _x2;

      return isNum$1(x) ? (_x = x, soleChalk(_x)) : (_x2 = x, naChalk(_x2));
    });
  }

  const rto = mapAr$1(dif, _ => _ / d, 3);
  return retFn ? map(ar, x => isNum$1(x) ? toHatsu(x, m, rto, min) : naChalk) : map(ar, x => {
    var _x3, _x4;

    return isNum$1(x) ? (_x3 = x, toHatsu(x, m, rto, min)(_x3)) : (_x4 = x, naChalk(_x4));
  });
};
/**
 *
 * @param {*[][]} mx
 * @param {number|*} m
 * @param {number|*} d
 * @param {[number,number,number]} min
 * @param {[number,number,number]} dif
 * @param {function} naChalk
 * @param {boolean} retFn
 * @param {boolean} mutate
 * @returns {function[]|*[]}
 */


const mxVisual$1 = (mx, {
  val: {
    min: m,
    dif: d
  },
  pal: {
    min,
    dif
  },
  naChalk,
  retFn,
  mutate
}) => {
  const map = mutate ? mutateMap$1 : mapAr$1;

  if (isNaN(d) || !d || dif.every(el => !el)) {
    var _min;

    const soleChalk = (_min = min, tube$1(_min));
    return retFn ? map(mx, el => isNum$1(el) ? soleChalk : naChalk) : map(mx, el => {
      var _el, _el2;

      return isNum$1(el) ? (_el = el, soleChalk(_el)) : (_el2 = el, naChalk(_el2));
    });
  }

  const rto = mapAr$1(dif, _ => _ / d, 3);
  return retFn ? map(mx, el => isNum$1(el) ? toHatsu(el, m, rto, min) : naChalk) : map(mx, el => {
    var _el3, _el4;

    return isNum$1(el) ? (_el3 = el, toHatsu(el, m, rto, min)(_el3)) : (_el4 = el, naChalk(_el4));
  });
};
/**
 *
 * @param {*[][]} mx
 * @param {number} y
 * @param {number|*} m
 * @param {number|*} d
 * @param {[number,number,number]} min
 * @param {[number,number,number]} dif
 * @param {function} naChalk
 * @param {boolean} retFn
 * @param {boolean} mutate
 * @returns {function[]|*[]}
 */


const mxVisualCol$1 = (mx, y, {
  val: {
    min: m,
    dif: d
  },
  pal: {
    min,
    dif
  },
  naChalk,
  retFn,
  mutate
}) => {
  const mapCol = mutate ? veho.Mx.mutateCol : veho.Mx.mapCol;

  if (isNaN(d) || !d || dif.every(x => !x)) {
    var _min2;

    const soleChalk = (_min2 = min, tube$1(_min2));
    return retFn ? mapCol(mx, y, el => isNum$1(el) ? soleChalk : naChalk) : mapCol(mx, y, el => {
      var _el5, _el6;

      return isNum$1(el) ? (_el5 = el, soleChalk(_el5)) : (_el6 = el, naChalk(_el6));
    });
  }

  const rto = mapAr$1(dif, _ => _ / d, 3);
  return retFn ? mapCol(mx, y, el => isNum$1(el) ? toHatsu(el, m, rto, min) : naChalk) : mapCol(mx, y, el => {
    var _el7, _el8;

    return isNum$1(el) ? (_el7 = el, toHatsu(el, m, rto, min)(_el7)) : (_el8 = el, naChalk(_el8));
  });
};

const {
  fromHex: fromHex$1,
  toHsl: toHsl$1
} = farbe.Rgb;

const parseHsl$1 = color => {
  var _ref, _color;

  return typeof color === 'string' ? (_ref = (_color = color, fromHex$1(_color)), toHsl$1(_ref)) : color;
};
/**
 *
 * @param max
 * @param min
 * @returns {{dif: [number,number,number], min: [number,number,number]}}
 */


const colorBound$1 = (max, min) => ({
  min,
  dif: veho.Ar.mutateMap(max, (a, i) => a - min[i], 3)
});

class Visual$1 {
  /**
   *
   * @param {*[]} arr
   * @param {string|number[]} [max]
   * @param {string|number[]} [min]
   * @param {string|number[]} [na]
   * @param {boolean} [mutate=true]
   * @param {boolean} [retFn=false]
   */
  static vector(arr, {
    mark: {
      max = Presets.fresh.max,
      min = Presets.fresh.min,
      na = Presets.fresh.na
    } = {},
    mutate = false,
    retFn = false
  } = {}) {
    var _max, _min, _ref2, _na;

    return vecVisual$1(arr, {
      val: borel.Stat.bound(arr, {
        dif: true,
        level: 2
      }),
      pal: colorBound$1((_max = max, parseHsl$1(_max)), (_min = min, parseHsl$1(_min))),
      naChalk: (_ref2 = (_na = na, parseHsl$1(_na)), tube$1(_ref2)),
      mutate,
      retFn
    });
  }
  /**
   *
   * @param {*[][]} mx
   * @param {string|number[]} [max]
   * @param {string|number[]} [min]
   * @param {string|number[]} [na]
   * @param {number} [direct=1] - 0:p-wise, 1:r-wise, 2:c-wise
   * @param {boolean} [mutate=true]
   * @param {boolean} [retFn=false]
   * @returns {*}
   */


  static matrix(mx, {
    mark: {
      max = Presets.subtle.max,
      min = Presets.subtle.min,
      na = Presets.subtle.na
    } = {},
    direct = 1,
    mutate = false,
    retFn = false
  } = {}) {
    var _max2, _min2, _ref3, _na2;

    if (!veho.Mx.isMat(mx)) throw new Error('Not valid matrix');

    switch (direct) {
      case 2:
        const mapCols = veho.Mx.mapColumns;
        return mapCols(mx, c => Visual$1.vector(c, {
          mark: {
            max,
            min,
            na
          },
          mutate,
          retFn
        }));

      case 1:
        const mapRows = mutate ? mutateMap$1 : mapAr$1;
        return mapRows(mx, r => Visual$1.vector(r, {
          mark: {
            max,
            min,
            na
          },
          mutate,
          retFn
        }));

      case 0:
      default:
        return mxVisual$1(mx, {
          val: borel.StatMx.bound(mx, {
            dif: true,
            level: 2
          }),
          pal: colorBound$1((_max2 = max, parseHsl$1(_max2)), (_min2 = min, parseHsl$1(_min2))),
          naChalk: (_ref3 = (_na2 = na, parseHsl$1(_na2)), tube$1(_ref3)),
          mutate,
          retFn
        });
    }
  }
  /**
   *
   * @param {*[][]} mx
   * @param {number} y
   * @param {string|number[]} [max]
   * @param {string|number[]} [min]
   * @param {string|number[]} [na]
   * @param {boolean} deep
   */


  static column(mx, y, {
    mark: {
      max = Presets.fresh.max,
      min = Presets.fresh.min,
      na = Presets.fresh.na
    } = {},
    mutate = false,
    retFn = false
  } = {}) {
    var _max3, _min3, _ref4, _na3;

    return mxVisualCol$1(mx, y, {
      val: borel.StatMx.boundCol(mx, y, {
        dif: true,
        level: 2
      }),
      pal: colorBound$1((_max3 = max, parseHsl$1(_max3)), (_min3 = min, parseHsl$1(_min3))),
      naChalk: (_ref4 = (_na3 = na, parseHsl$1(_na3)), tube$1(_ref4)),
      mutate,
      retFn
    });
  }

}

const {
  isNumeric
} = typen.Num;
const rn = '\r\n';
const tb = '  ';
const aeu = '(Ø)';
const zhChars = {
  dash: '－',
  space: '　'
};

const lpad = (tx, pd, ansi = false, fill) => ansi && hasAnsi(tx) ? tx.padStart(tx.length + pd - stringLength_1(tx), fill) : tx.padStart(pd, fill);

const rpad = (tx, pd, ansi = false, fill) => ansi && hasAnsi(tx) ? tx.padEnd(tx.length + pd - stringLength_1(tx), fill) : tx.padEnd(pd, fill);

const numPad = (tx, ref, pd, ansi = false, fill) => isNumeric(ref) ? lpad(tx, pd, ansi, fill) : rpad(tx, pd, ansi, fill);

const isTab = c => c === '\t' || c === ' ';

const deNaTab = tx => {
  let i = 0;

  for (let {
    length
  } = tx; i < length; i++) if (!isTab(tx.charAt(i))) return i;

  return i;
};

const afterNaTab = tx => tx.substring(deNaTab(tx));

const endsBracs = tx => tx.endsWith(')') || tx.endsWith(']');
/**
 *
 * @param {*} x
 * @return {string}
 */


const totx = x => `${x}`;

const StrReg = {
  jv2py: /[A-Z]+|[0-9]+/g,
  py2jv: /[A-Za-z\d]+/g
};

class StrX {
  /**
   * Java-style expression -> Pythonic expression
   * @example 'TheWallstreetJournal2025WSJ' -> 'the wallstreet journal 2025 wsj'
   * @param {string} jvExp java-style expression
   * @param {string} delimiter
   * @returns {string} pythonic expression
   */
  static jv2py(jvExp, delimiter = ' ') {
    return jvExp.replace(StrReg.jv2py, it => delimiter + it.toLowerCase()).trim();
  }
  /**
   * Pythonic expression -> Java-style expression
   * @example 'THE_WALLSTREET_JOURNAL-2019.FOR.THE.FANS' -> 'theWallstreetJournal2019ForTheFans'
   * @param {string} pyExp python expression
   * @returns {string} java expression
   */


  static py2jv(pyExp) {
    const matches = pyExp.match(StrReg.py2jv);
    return matches ? matches[0].toLowerCase() + matches.slice(1).map(wd => wd[0].toUpperCase() + wd.slice(1).toLowerCase()).join() : pyExp;
  }

  static wL(tx = '') {
    console.log(tx);
  }

  static tag(label, item) {
    const i = deNaTab(label);
    let [key, text] = [endsBracs(label) ? label : `${label.substring(0, i)}[${label.substring(i)}]`, `${item}`];

    if (text.includes('\n')) {
      const t = ' '.repeat(i);
      text = (text.endsWith('}') || text.endsWith(']')) && !text.endsWith(']]') ? afterNaTab(text.split(rn).map(x => t + x).join(rn)) : ['', ...text.split(rn).map(x => t + tb + x), t].join(rn);
    }

    return `${key} (${text})`;
  }

  static narrow(tx, lb, rb) {
    const [li, ri] = [tx.indexOf(lb), tx.lastIndexOf(rb)];
    return li > 0 && ri > 0 ? tx.slice(li, ri + rb.length) : tx;
  }

  static narrowExclude(tx, lb, rb) {
    const [li, ri] = [tx.indexOf(lb), tx.lastIndexOf(rb)];
    return li && ri ? tx.slice(li + lb.length, ri) : tx;
  }

  static padStartAnsi(tx, len, fill) {
    return hasAnsi(tx) ? tx.padStart(tx.length + len - stringLength_1(tx), fill) : tx.padStart(len, fill);
  }

  static padEndAnsi(tx, len, fill) {
    return hasAnsi(tx) ? tx.padEnd(tx.length + len - stringLength_1(tx), fill) : tx.padEnd(len, fill);
  }
  /**
   * Return if a string contains Chinese character.
   * halfAng = str.match(/[\u0000-\u00ff]/g) || [] //半角
   * chinese = str.match(/[\u4e00-\u9fa5]/g) || [] //中文
   * fullAng = str.match(/[\uff00-\uffff]/g) || [] //全角
   * @param {string} str
   * @returns {boolean}
   */


  static hasChn(str) {
    return str.search(/[\u4e00-\u9fa5]|[\uff00-\uffff]/) !== -1;
  }
  /**
   * Half-angle string -> Full-angle string
   * 半角转化为全角
   * a.全角空格为12288，半角空格为32
   * b.其他字符半角(33-126)与全角(65281-65374)的对应关系是：均相差65248
   * @param {string} tx
   * @returns {string}
   * @constructor
   */


  static toFullAngle(tx) {
    let t = '';

    for (let c of tx) {
      const co = c.charCodeAt(0);
      t = co === 32 ? t + String.fromCharCode(12288) : co < 127 ? t + String.fromCharCode(co + 65248) : t + c;
    }

    return t;
  }
  /**
   * Full-angle string -> Half-angle string
   * 全角转换为半角
   * @param {string} tx
   * @returns {string}
   * @constructor
   */


  static toHalfAngle(tx) {
    let t = '';

    for (let c of tx) {
      const co = c.charCodeAt(0);
      t += co === 12288 ? String.fromCharCode(co - 12256) : 65280 < co && co < 65375 ? String.fromCharCode(co - 65248) : String.fromCharCode(co);
    }

    return t;
  }

  static indexNonTab(tx) {
    let i = 0;

    for (let {
      length
    } = tx; i < length; i++) if (!isTab(tx.charAt(i))) return i;

    return i;
  }

  static afterNonTab(tx) {
    return tx.substring(deNaTab(tx));
  }

} // if (co === 12288) {
//   t += String.fromCharCode(co - 12256)
// } else if (co > 65280 && co < 65375) {
//   t += String.fromCharCode(co - 65248)
// } else {
//   t += String.fromCharCode(co)
// }


class Preci {
  constructor(ar, head, tail, dash) {
    this.ar = ar;
    this.head = head;
    this.tail = tail;
    this.dash = dash;
  }

  reboot(ar) {
    this.ar = ar;
    return this;
  }

  copy(ar) {
    return new Preci(ar, this.head, this.tail, this.dash);
  }

  static fromArr(ar, head = 0, tail = 0) {
    let dash = true;
    if (!ar || !ar.length) [head, tail, dash] = [0, 0, false];
    const length = ar.length;
    if (!head || head >= length) [head, tail, dash] = [length, 0, false]; // if (!tail || tail >= length) [head, tail, dash] = [head, 0, true]
    // if (head + tail >= length) [head, tail, dash] = [head, -1, true]

    const ht = ar === null || ar === void 0 ? void 0 : ar.length,
          _ar = Array(ht);

    for (let i = 0; i < head; i++) _ar[i] = ar[i];

    for (let i = ht - tail; i < ht; i++) _ar[i] = ar[i];

    return new Preci(_ar, head, tail, dash);
  }

  map(fn, mutate = false) {
    return mutate ? this.reboot(this.ar.map(fn)) : this.copy(this.ar.map(fn));
  }

  toList(el) {
    const {
      ar,
      head,
      tail
    } = this,
          ht = ar.length,
          df = ht - (head + tail),
          _ar = ar.slice();

    this.dash && el ? _ar.splice(head, df, el) : _ar.splice(head, df);
    return _ar;
  }
  /**
   *
   * @param {function} [abstract]
   * @param {boolean} [mutate]
   * @return { Preci }
   */


  stringify(abstract, mutate = true) {
    const brief = abstract ? _ => String(abstract(_)) : totx;
    return this.map(brief, mutate);
  }

}

const isVisual = visual => visual !== false && (visual === null || visual === void 0 ? void 0 : visual.on) !== false;

const mapAr$2 = veho.Ar.map;
/**
 *
 * @param {Array<?string>} arr
 * @param ansi
 */

function maxLen(arr, ansi = false) {
  return ansi ? Math.max(...arr.map(x => !!x ? stringLength_1(x) : 0)) : Math.max(...arr.map(x => !!x ? x.length : 0));
}
/**
 *
 * @param {string[]} words
 * @param {*[]} raws
 * @param {Chalk[]} pals
 * @param {number[]|number} pads
 * @param {boolean=false} [ansi]
 * @param {number} len
 * @return {string[]}
 */


function vecPalPad(words, raws, pals, pads, ansi, len) {
  return mapAr$2(words, (tx, i) => {
    var _numPad;

    return _numPad = numPad(tx, raws[i], pads[i], ansi), pals[i](_numPad);
  }, len);
}
/**
 *
 * @param {string[]} words
 * @param {*[]} raws
 * @param {number[]} [pads]
 * @param {boolean=false} [ansi]
 * @param {number} len
 * @return {string[]}
 */


function vecPad(words, raws, pads, ansi, len) {
  return mapAr$2(words, (tx, i) => numPad(tx, raws[i], pads[i], ansi), len);
}
/**
 *
 * @param {*[]} ar
 * @param {number} hd - head(left) length
 * @param {number} ts - tail(right) start index
 * @param {number} l - array length
 * @returns {*[]}
 */


const cropAr = (ar, hd, ts, l) => {
  const _r = Array(l);

  for (--l; l >= ts; l--) _r[l] = ar[l];

  for (--hd; hd >= 0; hd--) _r[hd] = ar[hd];

  return _r;
};
/**
 *
 * @param {*[]} ar
 * @param {function(*)|function(*,number)} fn
 * @param {number} hd - head(left) length
 * @param {number} ts - tail(right) start index, ts=l-right
 * @param {number} l - array length
 * @returns {*[]}
 */


const cropMapAr = (ar, fn, hd, ts, l) => {
  const _r = Array(l);

  for (--l; l >= ts; l--) _r[l] = fn(ar[l], l);

  for (--hd; hd >= 0; hd--) _r[hd] = fn(ar[hd], hd);

  return _r;
};

const cropMx = (mx, [tp, bt], [lf, rt], [ht, wd]) => {
  const _mx = Array(ht),
        bs = ht - bt,
        rs = wd - rt;

  for (let i = 0; i < tp; i++) _mx[i] = cropAr(mx[i], lf, rs, wd);

  for (let i = bs; i < ht; i++) _mx[i] = cropAr(mx[i], lf, rs, wd);

  return _mx;
};

const cropRow = (row, i, fn, lf, rs, wd) => {
  const _r = Array(wd);

  for (--wd; wd >= rs; wd--) _r[wd] = fn(row[wd], i, wd);

  for (--lf; lf >= 0; lf--) _r[lf] = fn(row[lf], i, lf);

  return _r;
};

const cropMapMx = (mx, fn, [tp, bt], [lf, rt], [ht, wd]) => {
  const _mx = Array(ht),
        bs = ht - bt,
        rs = wd - rt;

  for (let i = 0; i < tp; i++) _mx[i] = cropRow(mx[i], i, fn, lf, rs, wd);

  for (let i = bs; i < ht; i++) _mx[i] = cropRow(mx[i], i, fn, lf, rs, wd);

  return _mx;
};

class PreX {
  constructor(mx, [top, bottom], [left, right], [height, width], [xDash, yDash]) {
    this.mx = mx;
    this.tb = [top, bottom];
    this.lr = [left, right];
    this.hw = [height, width];
    this.dash = [xDash, yDash];
  }
  /**
   *
   * @param {*[][]} mx
   * @returns {PreX}
   */


  reboot(mx) {
    this.mx = mx;
    return this;
  }
  /**
   *
   * @param {*[][]} mx
   * @returns {PreX}
   */


  copy(mx) {
    return new PreX(mx, this.tb, this.lr, this.hw, this.dash);
  }

  get height() {
    return this.tb[0] + this.tb[1];
  }

  get width() {
    return this.lr[0] + this.lr[1];
  }

  xHeight(elCn) {
    return Math.min(this.hw[0], this.height + (this.dash[0] && elCn ? elCn : 0));
  }

  xWidth(elCn) {
    return Math.min(this.hw[1], this.width + (this.dash[1] && elCn ? elCn : 0));
  }

  emptyRow(el) {
    const {
      hw: [, w],
      lr: [l, r]
    } = this;
    return cropMapAr(Array(w), () => el, l, w - r, w);
  }

  get voidSize() {
    const [height, width] = this.hw;
    return [height - this.height, width - this.width];
  }

  static fromMx(mx, [top = 0, bottom = 0], [left = 0, right = 0], size = [0, 0]) {
    // ({ top, bottom, left, right, size }) |> console.log
    size = size || veho.Mx.size(mx);
    let [height, width] = size,
        [xDash, yDash] = [true, true];
    if (!height || !width) [top, bottom, xDash, yDash] = [0, 0, false, false];
    if (!top || top >= height) [top, bottom, xDash] = [height, 0, false];
    if (!left || left >= width) [left, right, yDash] = [width, 0, false];

    const _mx = cropMx(mx, [top, bottom], [left, right], [height, width]);

    return new PreX(_mx, [top, bottom], [left, right], [height, width], [xDash, yDash]);
  }

  map(fn, mutate = true) {
    return mutate ? this.reboot(cropMapMx(this.mx, fn, this.tb, this.lr, this.hw)) : this.copy(cropMapMx(this.mx, fn, this.tb, this.lr, this.hw)); // return mutate
    //   ? this.reboot(this.mx.map((r, i) => r.map((x, j) => fn(x, i, j))))
    //   : this.copy(this.mx.map((r, i) => r.map((x, j) => fn(x, i, j))))
  }

  toMx(el) {
    const {
      mx,
      tb: [top],
      lr: [left],
      dash: [xDash, yDash],
      voidSize: [dx, dy]
    } = this,
          _mx = veho.Mx.copy(mx);

    xDash && el ? _mx.splice(top, dx, this.emptyRow(el)) : _mx.splice(top, dx);

    if (yDash && el) {
      for (let row of _mx) row.splice(left, dy, el);
    } else {
      for (let row of _mx) row.splice(left, dy);
    }

    return _mx;
  }

  stringify(abstract, mutate = true) {
    const brief = abstract ? _ => String(abstract(_)) : totx;
    return this.map(brief, mutate);
  }

}

const destructPreX = (mx, [top, bottom], [left, right], {
  abstract,
  visual = {},
  ansi = false
}, [height, width]) => {
  var _visual;

  const prex = PreX.fromMx(mx, [top, bottom], [left, right], [height, width]),
        rawx = prex.toMx('..'),
        palx = (_visual = visual, isVisual(_visual)) ? Visual$1.matrix(rawx, { ...visual,
    retFn: true,
    mutate: false
  }) : null,
        wordx = prex.stringify(abstract).toMx('..');
  return {
    rawx,
    palx,
    wordx
  };
};

const padMx = (wordx, rawx, palx, pads, ansi) => {
  const [, wd] = veho.Mx.size(rawx);
  return palx ? wordx.map((words, i) => vecPalPad(words, rawx[i], palx[i], pads, ansi, wd)) : wordx.map((words, i) => vecPad(words, rawx[i], pads, ansi, wd));
};

const readCrop = ({
  head,
  tail
}) => [head, tail];

const {
  hasChn,
  toFullAngle
} = StrX;
/**
 *
 *
 * @param {*[][]} head
 * @param wordx
 * @param rawx
 * @param {Chalk[][]} [palx]
 * @param {boolean=false} [ansi]
 * @param {boolean=false} chinese
 * @return {{head: string[], blanc: string[], rows: string[][]}}
 */

const padTable = (head, wordx, rawx, palx, ansi = false, chinese = false) => {
  if (chinese) return padTableCn(head, wordx, rawx, palx, ansi);
  const pads = veho.Mx.columns([head].concat(wordx), col => maxLen(col, ansi)),
        [_head, blanc, rows] = [veho.Ar.zip(head, pads, (x, p) => lpad(x, p, ansi)), pads.map(l => '-'.repeat(l)), padMx(wordx, rawx, palx, pads, ansi)];
  return {
    head: _head,
    blanc,
    rows
  };
};
/**
 *
 *
 * @param {*[][]} head
 * @param wordx
 * @param rawx
 * @param {Chalk[][]} [palx]
 * @param {boolean=false} [ansi]
 * @return {{head: string[], blanc: string[], rows: string[][]}}
 */


const padTableCn = (head, wordx, rawx, palx, ansi = false) => {
  const {
    dash,
    space
  } = zhChars;
  /**
   *
   * @type {{pd:number,cn:boolean}[]}
   */

  const pads = veho.Mx.columns([head].concat(wordx), col => ({
    pd: maxLen(col, ansi),
    cn: col.some(hasChn)
  })),
        [_head, blanc, rows] = [veho.Ar.zip(head, pads, (x, {
    cn,
    pd
  }) => cn ? rpad(toFullAngle(x), pd, ansi, space) : rpad(x, pd, ansi)), pads.map(p => (p.cn ? dash : '-').repeat(p.pd)), palx ? wordx.map((words, i) => veho.Ar.zip(words, pads, (tx, {
    cn,
    pd
  }, j) => {
    var _numPad, _numPad2;

    return cn ? (_numPad = numPad(toFullAngle(tx), rawx[i][j], pd, ansi, space), palx[i][j](_numPad)) : (_numPad2 = numPad(tx, rawx[i][j], pd, ansi), palx[i][j](_numPad2));
  })) : wordx.map((words, i) => veho.Ar.zip(words, pads, (tx, {
    cn,
    pd
  }, j) => cn ? numPad(toFullAngle(tx), rawx[i][j], pd, ansi, space) : numPad(tx, rawx[i][j], pd, ansi)))];
  return {
    head: _head,
    blanc,
    rows
  };
};

class TableX {
  /**
   *
   * @param {
   *          {banner:*[],matrix:*[][],[title]:string,[types]:*[]} |
   *          {head:*[],rows:*[][],[title]:string,[types]:*[]} |
   *          {header:*[],rowSet:*[][],[title]:string,[types]:*[]}
   *        } table
   * @param {?function(*):string} [abstract]
   * @param {{[abstract]:?function(*):string,[head]:?number,[tail]:?number}} [_head]
   * @param {{[head]:?number,[tail]:?number}} [_rows]
   * @param {{
   *          [on]:boolean,
   *          [mark]:{
   *            [max]:string|number[],
   *            [min]:string|number[],
   *            [na]:string|number[],
   *          },
   *          [direct]:number
   *         }} [visual]
   * @param {boolean} [ansi=false]
   * @param {boolean} [chinese=false]
   * @return {string}
   */
  static brief(table, {
    abstract,
    head: _head = {
      abstract: null,
      head: 0,
      tail: 0
    },
    rows: _rows = {
      head: 0,
      tail: 0
    },
    visual = {
      on: true,
      mark: {
        max: palett.Palett.lightGreen.accent_3,
        min: palett.Palett.orange.accent_2,
        na: palett.Greys.blueGrey.lighten_3
      },
      direct: 2
    },
    ansi = false,
    chinese = false
  } = {}) {
    var _visual, _rows2, _head2;

    let head = table.head || table.banner || table.header,
        rows = table.rows || table.matrix || table.rowSet,
        blanc;
    const [ht, wd] = veho.Mx.size(rows);
    if (!ht || !wd) return aeu;
    const visualOn = (_visual = visual, isVisual(_visual));
    ansi = visualOn ? true : ansi;
    const hs = Preci.fromArr(head, _head.head, _head.tail).stringify(abstract).toList('..'),
          {
      rawx,
      palx,
      wordx
    } = destructPreX(rows, (_rows2 = _rows, readCrop(_rows2)), (_head2 = _head, readCrop(_head2)), {
      abstract,
      visual,
      ansi
    }, [ht, wd]);
    ({
      head,
      blanc,
      rows
    } = padTable(hs, wordx, rawx, palx, ansi, chinese));
    return [head.join(' | '), blanc.join('-+-')].concat(rows.map(row => row.join(' | '))).join(rn);
  }

}

const {
  hasChn: hasChn$1,
  toFullAngle: toFullAngle$1
} = StrX;

class CrosTabX {
  /**
   *
   * @param {{side:*[],banner:*[],matrix:*[][],[title]:string}} crosTab
   * @param {?function(*):string} [abstract]
   * @param {{[abstract]:?function(*):string,[head]:?number,[tail]:?number}} [_s]
   * @param {{[abstract]:?function(*):string,[head]:?number,[tail]:?number}} [_b]
   * @param {{
   *          [on]:boolean,
   *          [mark]:{
   *            [max]:string|number[],
   *            [min]:string|number[],
   *            [na]:string|number[],
   *          },
   *          [direct]:number
   *         }} [visual]
   * @param {boolean} [ansi=false]
   * @param {boolean} [chinese=false]
   * @return {string}
   */
  static brief(crosTab, {
    abstract,
    side: _s = {
      abstract,
      head: 0,
      tail: 0
    },
    banner: _b = {
      abstract,
      head: 0,
      tail: 0
    },
    visual = {
      on: true,
      mark: {
        max: palett.Greys.grey.lighten_5,
        min: palett.Greys.grey.darken_1,
        na: palett.Palett.indigo.lighten_2
      },
      direct: 2
    },
    ansi = false,
    chinese = false
  } = {}) {
    var _visual, _s2, _b2;

    let {
      side,
      banner,
      matrix
    } = crosTab,
        title,
        cue;
    const [ht, wd] = veho.Mx.size(matrix);
    if (!ht || !wd) return aeu;
    const visualOn = (_visual = visual, isVisual(_visual));
    ansi = visualOn ? true : ansi;
    side = Preci.fromArr(side, _s.head, _s.tail).stringify(_s.abstract).toList('..');
    banner = Preci.fromArr(banner, _b.head, _b.tail).stringify(_b.abstract).toList('..');
    const {
      rawx,
      palx,
      wordx
    } = destructPreX(matrix, (_s2 = _s, readCrop(_s2)), (_b2 = _b, readCrop(_b2)), {
      abstract,
      visual,
      ansi
    }, [ht, wd]);
    ({
      title,
      cue,
      side
    } = padSide(side, crosTab.title || '', ansi, chinese));
    const {
      head,
      blanc,
      rows
    } = padTable(banner, wordx, rawx, palx, ansi, chinese);
    head.unshift(title);
    blanc.unshift(cue);
    veho.Ar.zip(side, rows, (s, row) => row.unshift(s));
    return [head.join(' | '), blanc.join('-+-')].concat(rows.map(row => row.join(' | '))).join(rn);
  }

}

const padSide = (side, title, ansi, chinese) => {
  if (chinese) return padSideCn(side, title, ansi);
  const ts = [title].concat(side),
        pad = maxLen(ts, ansi),
        cue = '-'.repeat(pad);
  title = rpad(title, pad, ansi);
  side = mapAr$2(side, x => lpad(x, pad, ansi));
  return {
    title,
    cue,
    side
  };
};

const padSideCn = (side, title, ansi) => {
  const {
    dash,
    space
  } = zhChars,
        ts = [title].concat(side),
        pad = maxLen(ts, ansi),
        cn = ts.some(hasChn$1);

  if (cn) {
    title = rpad(toFullAngle$1(title), pad, ansi, space);
    const cue = dash.repeat(pad);
    side = mapAr$2(side, x => lpad(toFullAngle$1(x), pad, ansi, space));
    return {
      title,
      cue,
      side
    };
  } else {
    return padSide(side, title, ansi);
  }
};

const {
  hex
} = hatsu.Hatsu;
/**
 *
 * @type {{str: function, brc: function, udf: function, num: function, brk: function, idx: function}}
 */

const Pal$1 = {
  idx: hex(palett.Greys.brown.lighten_4),
  str: hex(palett.Palett.lightGreen.accent_2),
  num: hex(palett.Palett.deepOrange.accent_2),
  udf: hex(palett.Palett.deepPurple.accent_2),
  brk: hex(palett.Palett.blue.accent_2),
  brc: hex(palett.Palett.amber.base),
  fnc: hex(palett.Palett.green.accent_4)
};
const brkL = Pal$1.brk('[ '),
      brkR = Pal$1.brk(' ]'),
      brcL = Pal$1.brc('{ '),
      brcR = Pal$1.brc(' }');

const {
  hex: hex$1
} = hatsu.Hatsu;
const Thm = {
  brm: hex$1(palett.Palett.blueGrey.base),
  br: hex$1(palett.Palett.orange.lighten_3),
  pr: hex$1(palett.Palett.indigo.lighten_1)
};

String.prototype.wL = function () {
  console.log(this);
};

String.prototype.tag = function (val) {
  return StrX.tag(this, val);
};

Number.prototype.tag = function (val) {
  return StrX.tag(`${this}`, val);
};

function _defineProperty$1(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _classPrivateMethodGet(receiver, privateSet, fn) {
  if (!privateSet.has(receiver)) {
    throw new TypeError("attempted to get private field on non-instance");
  }

  return fn;
}
/**
 * If y >= 0 then sort rows, else (e.g. y===undefined) sort keys.
 * @param {*[]} keys
 * @param {*[][]} rows
 * @param {function(*,*):number} comparer
 * @param {number} [y]
 * @returns {[*[], *[][]]}
 */


const sortAlong = (keys, rows, comparer, y) => {
  const comp = (a, b) => comparer(a[0], b[0]),
        ks = Array(keys.length),
        rs = y >= 0 ? rows.map((r, i) => [r[y], keys[i], r]).sort(comp).map(([, k, r], i) => {
    ks[i] = k;
    return r;
  }) : rows.map((r, i) => [keys[i], r]).sort(comp).map(([k, r], i) => {
    ks[i] = k;
    return r;
  });

  return [ks, rs];
};

class CrosTab {
  /**
   *
   * @param {*[]} side
   * @param {*[]} banner
   * @param {*[][]} matrix
   * @param {string} [title]
   */
  constructor(side, banner, matrix, title) {
    _defineProperty$1(this, "side", void 0);

    _defineProperty$1(this, "banner", void 0);

    _defineProperty$1(this, "matrix", void 0);

    _defineProperty$1(this, "title", void 0);

    this.side = side;
    this.banner = banner;
    this.matrix = matrix;
    this.title = title || '';
  }
  /**
   * Shallow copy
   * @param {*[]} side
   * @param {*[]} banner
   * @param {*[][]} matrix
   * @param {string} [title]
   * @return {CrosTab}
   */


  static from({
    side,
    banner,
    matrix,
    title
  }) {
    return new CrosTab(side, banner, matrix, title);
  }
  /**
   * Shallow copy
   * @param {*[]} side
   * @param {*[]} banner
   * @param {function(number,number):*} func
   * @param {string} [title]
   * @return {CrosTab}
   */


  static ini({
    side,
    banner,
    func,
    title
  }) {
    const matrix = veho.Mx.ini(side === null || side === void 0 ? void 0 : side.length, banner === null || banner === void 0 ? void 0 : banner.length, (x, y) => func(x, y));
    return CrosTab.from({
      side,
      banner,
      matrix,
      title
    });
  }
  /**
   *
   * @param {*[]|boolean} [side]
   * @param {*[]|boolean} [banner]
   * @returns {*}
   */


  toSamples({
    side,
    banner
  }) {
    const [sideIsArr, bannerIsArr] = [Array.isArray(side) && side.length || side === true, Array.isArray(banner) && banner.length || banner === true];

    if (sideIsArr) {
      if (bannerIsArr) {
        var _side$map;

        const entries = banner.map(x => [x, this.coin(x)]);
        return (_side$map = side.map(x => this.matrix[this.roin(x)]), veho.Mx.transpose(_side$map)).map(col => veho.Ob.fromEntries(entries, i => col[i]));
      } else {
        var _this$matrix;

        const entries = side.map(x => [x, this.roin(x)]);
        return (_this$matrix = this.matrix, veho.Mx.transpose(_this$matrix)).map(col => veho.Ob.fromEntries(entries, i => col[i]));
      }
    } else {
      if (bannerIsArr) {
        const entries = banner.map(x => [x, this.coin(x)]);
        return this.matrix.map(row => veho.Ob.fromEntries(entries, i => row[i]));
      } else {
        return veho.Samples.fromCrosTab(this);
      }
    }
  }

  get toJson() {
    var _this$matrix2;

    return {
      side: this.side.slice(),
      banner: this.banner.slice(),
      matrix: (_this$matrix2 = this.matrix, veho.clone(_this$matrix2)),
      title: this.title
    };
  }

  get columns() {
    return veho.Mx.transpose(this.matrix);
  }
  /**
   *
   * @param {*[][]} [newMatrix]
   * @param {*[]} [newSide]
   * @param {*[]} [newBanner]
   * @returns {CrosTab}
   */


  reboot(newMatrix, newSide, newBanner) {
    if (newMatrix) this.matrix = newMatrix;
    if (newSide) this.side = newSide;
    if (newBanner) this.banner = newBanner;
    return this;
  }
  /**
   * Shallow copy
   * @param {*[][]} [newMatrix]
   * @param {*[]} [newSide]
   * @param {*[]} [newBanner]
   * @return {CrosTab}
   */


  clone(newMatrix, newSide, newBanner) {
    return new CrosTab(newSide || this.side.slice(), newBanner || this.banner.slice(), newMatrix || veho.Mx.clone(this.matrix), this.title);
  }

  get size() {
    return veho.Mx.size(this.matrix);
  }

  get ht() {
    var _this$side;

    return (_this$side = this.side) === null || _this$side === void 0 ? void 0 : _this$side.length;
  }

  get wd() {
    var _this$banner;

    return (_this$banner = this.banner) === null || _this$banner === void 0 ? void 0 : _this$banner.length;
  }

  cell(x, y) {
    return this.matrix[x][y];
  }

  queryCell(r, c) {
    const x = this.side.indexOf(r),
          y = this.banner.indexOf(c);
    return x < 0 || y < 0 ? null : this.matrix[x][y];
  }

  queryCoordinate(r, c) {
    return {
      x: this.side.indexOf(r),
      y: this.banner.indexOf(c)
    };
  }

  roin(r) {
    return Number.isInteger(r) && 0 <= r && r < this.ht ? r : this.side.indexOf(r);
  }

  coin(c) {
    return Number.isInteger(c) && 0 <= c && c < this.wd ? c : this.banner.indexOf(c);
  }

  row(r) {
    const x = this.roin(r);
    return this.matrix[x];
  }

  column(c) {
    const y = this.coin(c);
    return this.matrix.map(row => row[y]);
  }

  setRow(r, newRow) {
    const x = this.roin(r);
    this.matrix[x] = newRow;
    return this;
  }

  setColumn(c, newColumn) {
    const y = this.coin(c);
    this.matrix.forEach((row, i) => {
      row[y] = newColumn[i];
    });
    return this;
  }

  map(fn, {
    mutate = true
  } = {}) {
    const matrix = veho.Mx.map(this.matrix, fn);
    return mutate ? this.reboot(matrix) : this.clone(matrix);
  }
  /**
   * Push row into this and return this.
   * No shallow nor deep copy of the matrix.
   * @param {*} field
   * @param {*[]} row
   * @return {CrosTab}
   */


  pushRow(field, row) {
    this.side.push(field);
    this.matrix.push(row);
    return this;
  }
  /**
   * Unshift row into this and return this.
   * No shallow nor deep copy of the matrix.
   * @param {*} field
   * @param {*[]} row
   * @return {CrosTab}
   */


  unshiftRow(field, row) {
    this.side.unshift(field);
    this.matrix.unshift(row);
    return this;
  }
  /**
   * Push column into this and return this.
   * No shallow nor deep copy of the matrix.
   * @param {*} field
   * @param {*[]} col
   * @return {CrosTab}
   */


  pushCol(field, col) {
    this.banner.push(field);
    this.matrix.forEach((row, i) => row.push(col[i]));
    return this;
  }
  /**
   * Unshift column into this and return this.
   * No shallow nor deep copy of the matrix.
   * @param {*} field
   * @param {*[]} col
   * @return {CrosTab}
   */


  unshiftCol(field, col) {
    this.banner.unshift(field);
    this.matrix.forEach((row, i) => row.unshift(col[i]));
    return this;
  }

  select({
    side: s,
    banner: b
  } = {}, {
    mutate = true
  } = {}) {
    let {
      matrix,
      side,
      banner
    } = this;

    if (s && s.length) {
      const roins = s.map(x => this.roin(x));
      matrix = veho.Ar.select(matrix, roins);
      side = s;
    }

    if (b && b.length) {
      const coins = b.map(x => this.coin(x));
      matrix = veho.Mx.select(matrix, coins);
      banner = b;
    }

    return mutate ? this.reboot(matrix, side, banner) : this.clone(matrix, side, banner);
  }

  slice({
    rows = {
      begin: undefined,
      end: undefined
    },
    cols = {
      begin: undefined,
      end: undefined
    }
  }, {
    mutate = true
  } = {}) {
    let {
      side,
      banner,
      matrix
    } = this;
    let begin, end;

    if (rows) {
      ({
        begin,
        end
      } = rows);
      side = side.slice(begin, end);
      matrix = matrix.slice(begin, end);
    }

    if (cols) {
      ({
        begin,
        end
      } = cols);
      banner = this.banner.slice(begin, end);
      matrix = matrix.map(row => row.slice(begin, end));
    }

    return mutate ? this.reboot(matrix, side, banner) : this.clone(matrix, side, banner);
  }

  sort(by = 'rows', field, comparer, {
    mutate = true
  } = {}) {
    var _matrix, _matrix2;

    let {
      side,
      banner,
      matrix
    } = this;

    switch (by.charAt(0)) {
      case 'c':
        const x = this.roin(field);
        [banner, matrix] = sortAlong(banner, (_matrix = matrix, veho.Mx.transpose(_matrix)), comparer, x);
        matrix = (_matrix2 = matrix, veho.Mx.transpose(_matrix2));
        break;

      case 'r':
      default:
        const y = this.coin(field);
        [side, matrix] = sortAlong(side, matrix, comparer, y);
        break;
    }

    return mutate ? this.reboot(matrix, side, banner) : this.clone(matrix, side, banner);
  }

  sortLabel(by = 'rows', comparer) {
    var _matrix3, _matrix4;

    let {
      side,
      banner,
      matrix,
      title
    } = this;

    switch (by.charAt(0)) {
      case 'c':
        side = side.slice();
        [banner, matrix] = sortAlong(banner, (_matrix3 = matrix, veho.Mx.transpose(_matrix3)), comparer);
        matrix = (_matrix4 = matrix, veho.Mx.transpose(_matrix4));
        break;

      case 'r':
      default:
        banner = banner.slice();
        [side, matrix] = sortAlong(side, matrix, comparer);
        break;
    }

    return new CrosTab(side, banner, matrix, title);
  }

  transpose(newTitle, {
    mutate = true
  } = {}) {
    const {
      banner: side,
      side: banner,
      columns: matrix
    } = this;
    return mutate ? this.reboot(matrix, side, banner) : this.clone(matrix, side, banner);
  }

}

const {
  inferData
} = typen.Num;

class ArrTyp {
  /**
   *
   * @param {*[]} column
   * @return {string|unknown}
   */
  static infer(column) {
    if (column.length) {
      const types = column.map(inferData);
      const dist = new Set(types);

      switch (dist.size) {
        case 1:
          return types[0];

        case 2:
          return dist.has('number') && dist.has('numstr') ? 'numstr' : 'misc';

        default:
          return 'misc';
      }
    } else {
      return 'null';
    }
  }

}

const pivotMode = stat => {
  return typeof stat === 'string' ? veho.PivotModes[stat] || (stat === 'cnt' ? veho.PivotModes.count : veho.PivotModes.array) : veho.PivotModes.array;
};

veho.Pivot.prototype.pivotStat = function ([x, y], [index, stat], {
  boot
} = {}) {
  var _stat;

  const mode = (_stat = stat, pivotMode(_stat));
  let {
    side,
    banner,
    matrix
  } = this.pivot([x, y, index], {
    mode,
    boot
  });
  if (mode === veho.PivotModes.array) matrix = veho.Mx.map(matrix, stat);
  return {
    side,
    banner,
    matrix
  };
};
/**
 *
 * @param {number} x
 * @param {number} y
 * @param {{index:number,stat:string|function(*[]):number}[]}fields
 */


veho.Pivot.prototype.pivotMatrices = function ([x, y], fields) {
  const hi = fields.length;
  let {
    side,
    banner,
    matrix
  } = this.pivotStat([x, y], fields[0], {
    boot: true
  });
  const matrices = [matrix];

  for (let i = 1; i < hi; i++) {
    var _this$pivotStat$matri;

    _this$pivotStat$matri = this.pivotStat([x, y], fields[i], {
      boot: false
    }).matrix, matrices.push(_this$pivotStat$matri);
  }

  return {
    side,
    banner,
    matrix: veho.Mx.zip(matrices)
  };
};
/**
 *
 * @param {Table} table
 * @param {TableSpec} spec
 * @param {string} spec.side side
 * @param {string} spec.banner banner
 * @param {Object<string|number,function(*):boolean>} spec.filter
 * @param {string|string[]|Object<string|number,function(*[]):number>} spec.cell
 * @param {function():number} spec.calc - (field1,field2,...)=>number
 * @returns {CrosTab}
 */


const _crosTabFut = (table, spec) => {
  var _spec$side, _spec$banner, _field; // spec |> console.log


  table = table.filter(spec.filter, {
    mutate: false
  });
  const pvt = new veho.Pivot(table.matrix),
        coin = table.coin.bind(table),
        cellEntries = Object.entries(spec.cell),
        [x, y] = [(_spec$side = spec.side, coin(_spec$side)), (_spec$banner = spec.banner, coin(_spec$banner))];
  let side, banner, matrix;

  switch (cellEntries.length) {
    case 0:
      ({
        side,
        banner,
        matrix
      } = pvt.pivot([x, y], {
        mode: veho.PivotModes.count
      }));
      break;

    case 1:
      const [[field, stat]] = cellEntries;
      ({
        side,
        banner,
        matrix
      } = pvt.pivotStat([x, y], [(_field = field, coin(_field)), stat]));
      break;

    default:
      ({
        side,
        banner,
        matrix
      } = pvt.pivotMulti([x, y], cellEntries.map(([field, stat]) => {
        var _field2;

        return [(_field2 = field, coin(_field2)), stat];
      })));
  }

  const {
    calc
  } = spec;

  if (calc) {
    const fn = ar => calc.apply(null, ar);

    matrix = veho.Mx.map(matrix, fn);
  }

  return CrosTab.from({
    side,
    banner,
    matrix,
    spec: spec.title
  });
};
/**
 *
 * @type {{intersect: number, left: number, union: number, right: number}}
 */


const JoinT = {
  intersect: -1,
  union: 0,
  left: 1,
  right: 2
};
const {
  spliceCols,
  size
} = veho.Mx;
const {
  select,
  ini: ar
} = veho.Ar;

const locInd = (mx, ys, vs, hi) => mx.findIndex(row => {
  for (let i = 0; i < hi; i++) if (row[ys[i]] !== vs[i]) return false;

  return true;
});
/**
 *
 * @param joinT
 * @returns {
 * (function({mx: *[][], ys: number[]}, {mx: *[][], ys: number[]}): *[][])
 * |(function({mx: *[][], ys: number[]}, {mx: *[][], ys: number[]}, *=): *[][])
 * }
 */


const joiner = joinT => {
  switch (joinT) {
    case JoinT.union:
      return MxJoin.joinUnion;

    case JoinT.left:
      return MxJoin.joinLeft;

    case JoinT.right:
      return MxJoin.joinRight;

    case JoinT.intersect:
    default:
      return MxJoin.joinIntersect;
  }
};

class MxJoin {}

_defineProperty$1(MxJoin, "joinUnion", ({
  mx: mxL,
  ys: ysL
}, {
  mx: mxR,
  ys: ysR
}, fillEmpty) => {
  const hL = mxL.length,
        hR = mxR.length,
        hi = ysR.length,
        mx = Array(hL),
        mxL2 = spliceCols(veho.Mx.copy(mxL), ysL.slice().sort(borel.Comparer.numberAscending)),
        [, wL] = size(mxL2),
        mxR2 = spliceCols(veho.Mx.copy(mxR), ysR.slice().sort(borel.Comparer.numberAscending)),
        [, wR] = size(mxR2),
        idxRiLs = new Set();

  for (let i = 0, x, vsL; i < hL; i++) {
    vsL = select(mxL[i], ysL, hi);
    x = locInd(mxR, ysR, vsL, hi);

    if (x < 0) {
      mx[i] = vsL.concat(mxL2[i], ar(wR, fillEmpty));
    } else {
      mx[i] = vsL.concat(mxL2[i], mxR2[x]);
      idxRiLs.add(x);
    }
  }

  for (let i = 0, x, vsR; i < hR; i++) {
    var _vsR$concat;

    if (idxRiLs.has(i)) {
      continue;
    }

    vsR = select(mxR[i], ysR, hi);
    x = locInd(mxL, ysL, vsR, hi);
    if (x < 0) _vsR$concat = vsR.concat(ar(wL, fillEmpty), mxR2[i]), mx.push(_vsR$concat);
  }

  return mx;
});

_defineProperty$1(MxJoin, "joinLeft", ({
  mx: mxL,
  ys: ysL
}, {
  mx: mxR,
  ys: ysR
}, fillEmpty) => {
  const hL = mxL.length,
        hi = ysL.length,
        mx = Array(hL),
        mxL2 = spliceCols(veho.Mx.copy(mxL), ysL.slice().sort(borel.Comparer.numberAscending)),
        mxR2 = spliceCols(veho.Mx.copy(mxR), ysR.slice().sort(borel.Comparer.numberAscending)),
        [, wR] = size(mxR2);

  for (let i = 0, x, vsL; i < hL; i++) {
    vsL = select(mxL[i], ysL, hi);
    x = locInd(mxR, ysR, vsL, hi);
    mx[i] = x < 0 ? vsL.concat(mxL2[i], ar(wR, fillEmpty)) : vsL.concat(mxL2[i], mxR2[x]);
  }

  return mx;
});

_defineProperty$1(MxJoin, "joinRight", ({
  mx: mxL,
  ys: ysL
}, {
  mx: mxR,
  ys: ysR
}, fillEmpty) => {
  const hR = mxR.length,
        hi = ysR.length,
        mx = Array(hR),
        mxL2 = spliceCols(veho.Mx.copy(mxL), ysL.slice().sort(borel.Comparer.numberAscending)),
        [, wL] = size(mxL2),
        mxR2 = spliceCols(veho.Mx.copy(mxR), ysR.slice().sort(borel.Comparer.numberAscending));

  for (let i = 0, x, vsR; i < hR; i++) {
    vsR = select(mxR[i], ysR, hi);
    x = locInd(mxL, ysL, vsR, hi);
    mx[i] = x < 0 ? vsR.concat(ar(wL, fillEmpty), mxR2[i]) : vsR.concat(mxL2[x], mxR2[i]);
  }

  return mx;
});

_defineProperty$1(MxJoin, "joinIntersect", ({
  mx: mxL,
  ys: ysL
}, {
  mx: mxR,
  ys: ysR
}) => {
  const hL = mxL.length,
        hi = ysL.length,
        mx = [],
        mxL2 = spliceCols(veho.Mx.copy(mxL), ysL.slice().sort(borel.Comparer.numberAscending)),
        mxR2 = spliceCols(veho.Mx.copy(mxR), ysR.slice().sort(borel.Comparer.numberAscending));

  for (let i = 0, x, vsL; i < hL; i++) {
    vsL = select(mxL[i], ysL, hi);
    x = locInd(mxR, ysR, vsL, hi);
    if (x < 0) continue;
    mx.push(vsL.concat(mxL2[i], mxR2[x]));
  }

  return mx;
});

class TabQ {
  /**
   *
   * @param {Table} table
   * @param {Table} another
   * @param {string[]|number[]} indexFields
   * @param {number} [joinType=-1] - union:0,left:1,right:2,intersect:-1
   * @param {*} [fillEmpty]
   * @returns {Table}
   */
  static join(table, another, indexFields, joinType = JoinT.intersect, fillEmpty = null) {
    const ysL = indexFields.map(x => table.coin(x)),
          ysR = indexFields.map(x => another.coin(x)),
          joinFn = joiner(joinType),
          matrix = joinFn({
      mx: table.matrix,
      ys: ysL
    }, {
      mx: another.matrix,
      ys: ysR
    }, fillEmpty),
          banner = veho.Ar.select(table.banner, ysL).concat(veho.Ar.splices(table.banner.slice(), ysL.slice().sort(borel.Comparer.numberAscending)), veho.Ar.splices(another.banner.slice(), ysL.slice().sort(borel.Comparer.numberAscending)));
    return new Table(banner, matrix, `${table.title} ${another.title}`);
  }

}

class Table {
  /**
   *
   * @param {*[]} [banner]
   * @param {*[][]} [matrix]
   * @param {string} [title]
   * @param {*[]} [types]
   */
  constructor(_banner, matrix, title, types) {
    _fieldIndexes.add(this);

    this.banner = _banner || [];
    this.matrix = matrix || [[]];
    this.title = title || '';
    this.types = types;
  }
  /**
   *
   * @returns {[]|Array}
   */


  get types() {
    return this._types;
  }
  /**
   *
   * @param {Array} [value]
   */


  set types(value) {
    this._types = veho.Ar.isEmpty(value) ? [] : value;
  }
  /**
   *
   * @param {*[]} [banner]
   * @param {*[][]} [matrix]
   * @param {string} [title]
   * @param {string[]} [types]
   * @return {Table}
   */


  static from({
    banner,
    matrix,
    title,
    types
  }) {
    return new Table(banner, matrix, title, types);
  }
  /**
   *
   * @param {...string|number|[*,*]} fields
   * @returns {Object<*, *>[]}
   */


  toSamples(...fields) {
    if (fields.length) {
      const entries = fields.map(x => Array.isArray(x) ? [[x[1]], this.coin(x[0])] : [x, this.coin(x)]);
      return this.matrix.map(row => veho.Ob.fromEntries(entries, i => row[i]));
    }

    return this.matrix.map(row => veho.Ob.ini(this.banner, row));
  }
  /**
   *
   * @param {boolean} [mutate=true]
   * @returns {*}
   */


  toJson(mutate = true) {
    var _matrix;

    const {
      banner,
      matrix,
      title
    } = this;
    return mutate ? {
      banner,
      matrix,
      title
    } : {
      banner: banner.slice(),
      matrix: (_matrix = matrix, veho.Mx.clone(_matrix)),
      title: title.slice()
    };
  }
  /**
   *
   * @param {{}[]} samples
   * @param {?string} [title]
   * @param {?*[]} [types]
   * @param {*[]|[*,*][]} [fields]
   * @return {Table}
   */


  static fromSamples(samples, title, types, {
    fields
  } = {}) {
    const {
      head,
      rows
    } = veho.Samples.toTable(samples, {
      fields
    });
    return new Table(head, rows, title, types);
  }
  /**
   * Return 'this' by loading a new matrix
   * @param {*[][]} newMx
   * @param {*[]} [newBanner]
   * @param {string[]} [newTypes]
   * @return {Table}
   */


  reboot(newMx, newBanner, newTypes) {
    this.matrix = newMx;
    if (newBanner) this.banner = newBanner;
    if (newTypes) this.types = newTypes;
    return this;
  }
  /**
   *
   * @param {*[][]} [newMx]
   * @param {*[]} [newBanner]
   * @param {*[]} [newTypes]
   * @return {Table}
   */


  clone(newMx, newBanner, newTypes) {
    var _this$matrix;

    return new Table(newBanner || this.banner.slice(), newMx || (_this$matrix = this.matrix, veho.Mx.clone(_this$matrix)), this.title, newTypes || this.types.slice());
  }
  /**
   *
   * @param {*[]|[*,*][]} fields
   * @returns {{indexes: [], banner: []}|{indexes: *, banner: *}}
   */

  /**
   *
   * @param {*[]|[*,*][]} fields
   * @param {boolean=true} [mutate]
   * @returns {Table}
   */


  select(fields, {
    mutate = true
  } = {}) {
    const hi = fields === null || fields === void 0 ? void 0 : fields.length;
    if (!hi) return mutate ? this : this.clone();

    const {
      indexes,
      banner
    } = _classPrivateMethodGet(this, _fieldIndexes, _fieldIndexes2).call(this, fields);

    const matrix = veho.Mx.select(this.matrix, indexes);
    return mutate ? this.reboot(matrix, banner) : this.clone(matrix, banner);
  }
  /**
   *
   * @param {*[]|[*,*][]} fields
   * @param {boolean=true} [mutate]
   * @returns {Table}
   */


  spliceColumns(fields, {
    mutate = true
  } = {}) {
    var _matrix2;

    const ys = fields.map(c => this.coin(c)).sort(borel.Comparer.numberAscending),
          {
      matrix,
      banner
    } = this;
    return mutate ? this.reboot(veho.Mx.spliceCols(matrix, ys), veho.Ar.splices(banner, ys)) : this.clone(veho.Mx.spliceCols((_matrix2 = matrix, veho.Mx.copy(_matrix2)), ys), veho.Ar.splices(banner.slice(), ys));
  }

  map(fn, {
    mutate = true
  } = {}) {
    let {
      matrix
    } = this;
    matrix = veho.Mx.map(matrix, fn);
    return mutate ? this.reboot(matrix) : this.clone(matrix);
  }

  mapBanner(fn, {
    mutate = true
  } = {}) {
    let {
      banner,
      matrix
    } = this;
    banner = banner.map(fn);
    return mutate ? this.reboot(matrix, banner) : this.clone(matrix, banner);
  }
  /**
   *
   * @param {Table} another
   * @param {string[]|number[]} indexFields
   * @param {number} joinType - union:0,left:1,right:2,intersect:-1
   * @param {*} [fillEmpty]
   * @returns {Table}
   */


  join(another, indexFields, joinType = JoinT.intersect, fillEmpty = null) {
    return TabQ.join(this, another, indexFields, joinType, fillEmpty);
  }

  get size() {
    return veho.Mx.size(this.matrix);
  }

  get ht() {
    return !this.matrix ? 0 : this.matrix.length;
  }

  get wd() {
    return !this.banner ? 0 : this.banner.length;
  }

  get columns() {
    return veho.Mx.transpose(this.matrix);
  } // set types (value) {
  //
  // }


  cell(x, y) {
    return this.matrix[x][y];
  }
  /**
   *
   * @param {string|number} c
   */


  coin(c) {
    return Number.isInteger(c) && 0 <= c && c < this.wd ? c : this.banner.indexOf(c);
  }

  column(c) {
    const y = this.coin(c);
    return this.matrix.map(row => row[y]);
  }

  setColumn(c, newColumn) {
    const y = this.coin(c),
          {
      ht,
      matrix
    } = this;

    for (let i = 0; i < ht; i++) matrix[i][y] = newColumn[i];

    return this;
  }
  /**
   *
   * @param {string|number} c
   * @param {function} ject
   */


  setColumnBy(c, ject) {
    const y = this.coin(c);

    for (let row of this.matrix) row[y] = ject(row[y]);

    return this;
  }
  /**
   * Push row into this and return this.
   * No shallow nor deep copy of the matrix.
   * @param {*[]} row
   * @return {Table}
   */


  pushRow(row) {
    this.matrix.push(row);
    return this;
  }
  /**
   * Unshift row into this and return this.
   * No shallow nor deep copy of the matrix.
   * @param {*[]} row
   * @return {Table}
   */


  unshiftRow(row) {
    this.matrix.unshift(row);
    return this;
  }
  /**
   * Push column into this and return this.
   * No shallow nor deep copy of the matrix.
   * @param {*} name
   * @param {*[]} col
   * @return {Table}
   */


  pushCol(name, col) {
    this.banner.push(name);
    this.matrix.forEach((row, i) => row.push(col[i]));
    return this;
  }
  /**
   * Unshift column into this and return this.
   * No shallow nor deep copy of the matrix.
   * @param {*} name
   * @param {*[]} col
   * @return {Table}
   */


  unshiftCol(name, col) {
    this.banner.unshift(name);
    this.matrix.forEach((row, i) => row.unshift(col[i]));
    return this;
  }
  /**
   *
   * Specify the type of a column. No return
   * @param {string|number} field accept both column name in string or column index in integer
   * @param {string} typeName string | (number, float) | integer | boolean
   */


  changeType(field, typeName) {
    const y = this.coin(field);
    let column = veho.Mx.column(this.matrix, y);

    switch (typeName) {
      case 'string':
        column = column.map(totx);
        break;

      case 'number':
      case 'float':
        column = column.map(Number.parseFloat);
        break;

      case 'integer':
        column = column.map(Number.parseInt);
        break;

      case 'boolean':
        column = column.map(Boolean);
        break;

      default:
        typeName = this._types[y];
        break;
    }

    this.setColumn(field, column);
    this._types[y] = typeName;
    return this;
  }
  /**
   * Re-generate this._types based on DPTyp.inferArr method.
   * Cautious: This method will change all elements of this._types.
   * @return {string[]}
   */


  inferTypes() {
    const {
      infer
    } = ArrTyp;
    this.types = this.columns.map(infer);

    for (let [i, typ] of this.types.entries()) {
      switch (typ) {
        case 'numstr':
          this.changeType(i, 'number');
          break;

        case 'misc':
          this.changeType(i, 'string');
          break;
      }
    }

    return this.types;
  }
  /**
   *
   * @param {Object<string|number,function(*):boolean>} filters
   * @param {boolean} [mutate=true]
   * @return {Table}
   */


  filter(filters, {
    mutate = true
  } = {}) {
    let mx = this.matrix,
        y;

    for (let [field, crit] of Object.entries(filters)) {
      y = this.coin(field);
      if (y >= 0) mx = mx.filter(row => crit(row[y]));
    }

    return mutate ? this.reboot(mx) : this.clone(mx);
  }

  distinct(fields, {
    mutate = true
  } = {}) {
    let mx = this.matrix;

    for (let field of fields) mx = borel.StatMx.distinct(mx, this.coin(field));

    return mutate ? this.reboot(mx) : this.clone(mx);
  }
  /**
   *
   * @param {string|number} field
   * @param {boolean} [count=false]
   * @param {string|boolean} [sort=false] - When sort is function, sort must be a comparer between two point element.
   * @returns {[any, any][]|[]|any[]|*}
   */


  distinctCol(field, {
    count = false,
    sort = false
  } = {}) {
    const y = this.coin(field);
    return borel.StatMx.distinctCol(this.matrix, y, {
      count,
      sort
    });
  }

  sort(field, comparer, {
    mutate = true
  } = {}) {
    const y = this.coin(field),
          comp = (a, b) => comparer(a[y], b[y]),
          mx = this.matrix.slice().sort(comp);

    return mutate ? this.reboot(mx) : this.clone(mx);
  }
  /**
   *
   * @param {function(*,*):number} comparer - Comparer of banner elements
   * @param {boolean} mutate
   * @returns {Table|*}
   */


  sortLabel(comparer, {
    mutate = true
  } = {}) {
    var _columns, _columns2;

    let {
      banner,
      columns
    } = this;
    [banner, columns] = sortAlong(banner, columns, comparer);
    return mutate ? this.reboot((_columns = columns, veho.Mx.transpose(_columns)), banner) : this.clone((_columns2 = columns, veho.Mx.transpose(_columns2)), banner);
  }
  /**
   *
   * @param {TableSpec} spec
   * @param {string} spec.side side
   * @param {string} spec.banner banner
   * @param {{field:string, crit:function(*):boolean}[]} spec.filter
   * @param {{field:string, stat:function([]):number}[]} spec.cell
   * @param {function({}):number} spec.calc - ({col1,col2,...})=>number
   * @returns {CrosTab}
   */


  crosTab(spec) {
    return _crosTabFut(this, spec);
  }
  /**
   *
   * @param {TableSpec} spec
   * @param {string} spec.side side
   * @param {string} spec.banner banner
   * @param {{field:string, crit:function(*):boolean}[]} spec.filter
   * @param {{field:string, stat:function([]):number}[]} spec.cell
   * @param {function({}):number} spec.calc - ({col1,col2,...})=>number
   * @returns {CrosTab}
   */


  figure(spec) {
    return _crosTabFut(this, spec);
  }

}

var _fieldIndexes = new WeakSet();

var _fieldIndexes2 = function _fieldIndexes2(fields) {
  const hi = fields === null || fields === void 0 ? void 0 : fields.length;
  if (!hi) return {
    indexes: [],
    banner: []
  };
  const [indexes, banner] = [Array(hi), Array(hi)];

  for (let i = 0, x; i < hi; i++) {
    x = fields[i];
    [indexes[i], banner[i]] = Array.isArray(x) ? [this.coin(x[0]), x[1]] : [this.coin(x), x];
  }

  return {
    indexes,
    banner
  };
}; // Create an object type Er
/**
 *
 * @param {?function(*):string} [abstract]
 * @param {{[head]:?number,[tail]:?number}} [rows]
 * @param {{[abstract]:?function(*):string,[head]:?number,[tail]:?number}} [head]
 * @param {{
 *          [on]:boolean,
 *          [mark]:{
 *            [max]:string|number[],
 *            [min]:string|number[],
 *            [na]:string|number[],
 *          },
 *          [direct]:number
 *         }} [visual]
 * @param {boolean} [chinese=false]
 * @param {boolean} [ansi=false]
 * @return {string}
 */


Table.prototype.brief = function ({
  abstract = null,
  head = {
    head: 0,
    tail: 0
  },
  rows = {
    abstract: null,
    head: 0,
    tail: 0
  },
  visual = {
    on: true,
    mark: {
      max: palett.Palett.lightGreen.accent_3,
      min: palett.Palett.orange.accent_2,
      na: palett.Greys.blueGrey.lighten_3
    },
    direct: 2
  },
  chinese = false,
  ansi = false
} = {}) {
  return TableX.brief(this, {
    abstract,
    head,
    rows,
    visual,
    chinese,
    ansi
  });
};
/**
 *
 * @param {CrosTab|{side:*[],banner:*[],matrix:*[][],[title]:string}} crosTab
 * @param {function(*):string} [abstract]
 * @param {{[abstract]:?function(*):string,[head]:?number,[tail]:?number}} [side]
 * @param {{[abstract]:?function(*):string,[head]:?number,[tail]:?number}} [banner]
 * @param {{
 *          [on]:boolean,
 *          [mark]:{
 *            [max]:string|number[],
 *            [min]:string|number[],
 *            [na]:string|number[],
 *          },
 *          [direct]:number
 *         }} [visual]
 * @param {boolean} [chinese=false]
 * @param {boolean} [ansi=false]
 * @return {string}
 */


CrosTab.prototype.brief = function ({
  abstract = undefined,
  side = {
    abstract: undefined,
    head: 0,
    tail: 0
  },
  banner = {
    abstract: undefined,
    head: 0,
    tail: 0
  },
  visual = {
    on: true,
    mark: {
      max: palett.Greys.grey.lighten_5,
      min: palett.Greys.grey.darken_1,
      na: palett.Palett.indigo.lighten_2
    },
    direct: 2
  },
  chinese = false,
  ansi = false
} = {}) {
  return CrosTabX.brief(this, {
    abstract,
    side,
    banner,
    visual,
    chinese,
    ansi
  });
};

// const { random } = Math
const {
  random,
  abs,
  exp,
  log,
  sqrt,
  pow,
  cos,
  sin,
  PI
} = Math;

const randIdx = arr => ~~(random() * arr.length);

const randSel = arr => arr[randIdx(arr)];

const R0 = 3.442619855899;
const R1 = 1.0 / R0;
const R0S = exp(-0.5 * R0 * R0);
const N2P32 = -pow(2, 32);
const M1 = 2147483648.0;
const VN = 9.91256303526217e-3;

class Ziggurat {
  constructor(mean = 0, stdev = 1) {
    this.jsr = 123456789;
    this.wn = Array(128);
    this.fn = Array(128);
    this.kn = Array(128);
    this.mean = mean;
    this.stdev = stdev;
    this.preset();
  }

  preset() {
    // seed generator based on current time
    this.jsr ^= new Date().getTime();
    let m1 = M1,
        dn = R0,
        tn = R0,
        vn = VN,
        q = vn / R0S;
    this.kn[0] = ~~(dn / q * m1);
    this.kn[1] = 0;
    this.wn[0] = q / m1;
    this.wn[127] = dn / m1;
    this.fn[0] = 1.0;
    this.fn[127] = R0S;

    for (let i = 126; i >= 1; i--) {
      dn = sqrt(-2.0 * log(vn / dn + exp(-0.5 * dn * dn)));
      this.kn[i + 1] = ~~(dn / tn * m1);
      tn = dn;
      this.fn[i] = exp(-0.5 * dn * dn);
      this.wn[i] = dn / m1;
    }
  }

  next() {
    return this._randSample() * this.stdev + this.mean;
  }

  _randSample() {
    let hz = this._xorshift(),
        iz = hz & 127;

    return abs(hz) < this.kn[iz] ? hz * this.wn[iz] : this._nFix(hz, iz);
  }

  _nFix(hz, iz) {
    let r = R0,
        r1 = R1,
        x,
        y;

    while (true) {
      x = hz * this.wn[iz];

      if (iz === 0) {
        do {
          x = -log(this._uni()) * r1;
          y = -log(this._uni());
        } while (y + y < x * x); // {
        //   x = -log(this._uni()) * r1
        //   y = -log(this._uni())
        // }


        return hz > 0 ? r + x : -r - x;
      }

      if (this.fn[iz] + this._uni() * (this.fn[iz - 1] - this.fn[iz]) < exp(-0.5 * x * x)) return x;
      hz = this._xorshift();
      iz = hz & 127;
      if (abs(hz) < this.kn[iz]) return hz * this.wn[iz];
    }
  }

  _xorshift() {
    let m = this.jsr,
        n = m;
    n ^= n << 13;
    n ^= n >>> 17;
    n ^= n << 5;
    this.jsr = n;
    return m + n | 0;
  }

  _uni() {
    return 0.5 + this._xorshift() / N2P32;
  }

}

class Roulett {
  /**
   * Generate a random integer between [min, max].
   * Both min & max are inclusive.
   * @param {Number} min  Int
   * @param {Number} max  Int
   * @returns {Number}  Int
   */
  static between(min, max) {
    return ~~(random() * (max - min + 1)) + min;
  }
  /**
   * Generate a random integer between [min, max).
   * Notice: min is inclusive & max is exclusive.
   * @param {Number} min  Int
   * @param {Number} max(exclusive)  Int
   * @returns {Number}  Int
   */


  static rand(min, max) {
    return ~~(random() * (max - min)) + min;
  }

  static index(len) {
    return ~~(random() * len);
  }

  static arrayIndex(arr) {
    return randIdx(arr);
  }

  static element(arr) {
    return randSel(arr);
  }

  static key(obj) {
    return randSel(Object.keys(obj));
  }

  static value(obj) {
    return randSel(Object.values(obj));
  }

  static entry(obj) {
    return randSel(Object.entries(obj));
  }

  static gaussSamples(len, mean = 0, stdev = 1) {
    const z = new Ziggurat(mean, stdev),
          arr = Array(len);

    for (let i = 0; i < len; i++) arr[i] = z.next();

    return arr;
  }

}

const expReg = /[A-Z]+|[0-9]+/g;

const sepWords = jvExp => jvExp.replace(expReg, it => ' ' + it.toLowerCase()).trim();
/**
 *
 * @param {string} name
 * @returns {string}
 */


const readify = name => {
  var _name;

  return (_name = name, sepWords(_name)).replace('light ', 'l.').replace('deep ', 'd.');
};

const toDataPicker = space => {
  switch (space) {
    case 'rgb':
      return x => {
        var _x;

        return _x = x, farbe.Rgb.fromHex(_x);
      };

    case 'hsl':
      return x => {
        var _ref, _x2;

        return _ref = (_x2 = x, farbe.Rgb.fromHex(_x2)), farbe.Rgb.toHsl(_ref);
      };

    case 'hex':
    default:
      return x => x;
  }
};

function toStatistic(space, cellColor = false) {
  switch (space) {
    case 'rgb':
    case 'hsl':
      return arr => {
        let [a, b, c] = [0, 0, 0],
            len = arr.length;
        veho.Ar.map(arr, ([x, y, z]) => [a, b, c] = [a + x, b + y, c + z], len);
        return [borel.Zu.round(a / len), borel.Zu.round(b / len), borel.Zu.round(c / len)];
      };

    case 'hex':
      return arr => {
        var _ref2, _arr$map;

        return _ref2 = (_arr$map = arr.map(farbe.Rgb.fromHex), toStatistic('rgb')(_arr$map)), farbe.Rgb.toHex(_ref2);
      };

    default:
      return x => x;
  }
}

const toFormatter = (space, colorify = false) => {
  var _space;

  let formatter = space === 'rgb' || space === 'hsl' ? vec => vec.map(x => String(borel.Zu.round(x)).padStart(3)) : null;
  let hatsu = colorify ? (_space = space, toHatsu$1(_space)) : null;
  return formatter ? hatsu ? x => {
    var _ref3, _x3, _x4;

    return _ref3 = (_x3 = x, formatter(_x3)), (_x4 = x, hatsu(_x4))(_ref3);
  } : formatter : hatsu ? x => {
    var _x5, _x6;

    return _x5 = x, (_x6 = x, hatsu(_x6))(_x5);
  } : x => x;
};

function toHatsu$1(space) {
  switch (space) {
    case 'rgb':
      return rgb => hatsu.Hatsu.rgb(rgb).inverse;

    case 'hsl':
      return hsl => hatsu.Hatsu.hsl(hsl).inverse;

    case 'hex':
      return hex => hatsu.Hatsu.hex(hex).inverse;

    default:
      return null;
  }
}

const accents = veho.Ar.ini(4, i => `accent_${i + 1}`).reverse(),
      lightens = veho.Ar.ini(5, i => `lighten_${i + 1}`).reverse(),
      darkens = veho.Ar.ini(4, i => `darken_${i + 1}`);
const Degrees = {
  entire: [...accents, ...lightens, 'base', ...darkens],
  base: ['base'],
  lightens: lightens,
  darkens: darkens,
  accents: accents,
  readable: [...accents.slice(-3), ...lightens.slice(-3), 'base']
};
const red = ['red', 'pink'];
const purple = ['purple', 'deepPurple'];
const blue = ['indigo', 'blue', 'lightBlue', 'cyan'];
const green = ['teal', 'green'];
const yellowGreen = ['lightGreen', 'lime', 'yellow'];
const orange = ['amber', 'orange', 'deepOrange'];
const grey = ['brown', 'blueGrey', 'grey'];
const rainbow = [].concat(red, purple, blue, green, yellowGreen, orange);
const entire = rainbow.concat(grey);
const ColorGroups = {
  red,
  purple,
  blue,
  green,
  yellowGreen,
  orange,
  grey,
  rainbow,
  entire
};
const oc = Object.prototype.toString;
/**
 *
 * @param {*} o
 * @return {*}
 */

function clone(o) {
  if (!o || typeof o !== 'object') return o;

  switch (oc.call(o).slice(8, 11)) {
    case 'Arr':
      return dpArr(o);

    case 'Obj':
      return dpObj(o);

    case 'Map':
      return dpMap(o);

    case 'Dat':
      return new Date(+o);

    case 'Set':
      return new Set(dpArr([...o]));
  }

  throw new Error('Unable to copy obj. Unsupported type.');
}
/**
 *
 * @param {Map<*, *>} o
 * @return {Map<*, *>}
 */


function dpMap(o) {
  return new Map([...o.entries()].map(([k, v]) => [k, clone(v)]));
}
/**
 *
 * @param {*[]} o
 * @return {*[]}
 */


function dpArr(o) {
  return o.map(clone);
}
/**
 * Known issue:
 * Unable to clone circular and nested object.
 * @param {{}} o
 * @return {{}}
 */


function dpObj(o) {
  const x = {};

  for (let [k, v] of Object.entries(o)) x[k] = clone(v);

  return x;
} // let protoType = function (it) {
//   const raw = typeof it;
//   if (raw === "object") {
//     switch (true) {
//       case (it instanceof Array):
//         return "array";
//       case (it instanceof Map):
//         return "map";
//       case (it instanceof Set):
//         return "set";
//       case (it instanceof Function):
//         return "function";
//       default:
//         return raw;
//     }
//   } else {
//     return raw
//   }
// };


const oc$1 = Object.prototype.toString;
/**
 * const rxObj = /^\[object (.*)]$/
 * Equivalent to: Object.prototype.stringify.call(o).match(rxObj)[1]
 * @param {*} o
 * @return {string}
 */

const otype = o => oc$1.call(o).slice(8, -1);
/**
 * validate
 * @param x
 * @param y
 * @returns {number}
 */


const vdt = (x, y) => isNaN(x - y) ? NaN : y;

class Num {
  // Angular 4.3
  static isNumeric(x) {
    return !isNaN(x - parseFloat(x));
  }

  static numeric(x) {
    return vdt(x, parseFloat(x));
  }

  static inferData(x) {
    const t = typeof x;
    return t === 'string' ? isNaN(x - parseFloat(x)) ? 'string' : 'numstr' : t === 'object' ? otype(x).toLowerCase() : t;
  }

}

const check = x => !!x || x === 0;

class NumLoose {
  static isNumeric(x) {
    return check(+x);
  }

  static numeric(x) {
    x = +x;
    return check(x) ? x : NaN;
  }
  /**
   *
   * @param {*} x
   * @return {string}
   */


  static inferData(x) {
    const t = typeof x;
    return t === 'string' ? check(+x) ? 'numstr' : 'string' : t === 'object' ? otype(x).toLowerCase() : t;
  }

}
/**
 * Static class containing methods create 1d-array.
 */


const {
  numeric: num
} = Num,
      {
  numeric: numLoose
} = NumLoose;

class Ar {
  /**
   * Create an array.
   * @param {number} size Integer starts at zero.
   * @param {function(number):*|*} [ject] Defines the how index i decides value(i).
   * @returns {number[]} The
   */
  static ini(size, ject) {
    if (size <= 128) {
      let arr = [];

      if (typeof ject === 'function') {
        for (let i = 0; i < size; i++) arr[i] = ject(i);
      } else {
        for (let i = 0; i < size; i++) arr[i] = ject;
      }

      return arr;
    } else {
      return typeof ject === 'function' ? Array(size).fill(null).map((_, i) => ject(i)) : Array(size).fill(ject);
    }
  }

  static isEmpty(arr) {
    return !arr || !arr.length;
  }
  /**
   *
   * @param {Array} arr
   * @param {boolean} [loose]=false
   * @returns {*}
   */


  static numeric(arr, {
    loose = false
  }) {
    return arr.map(loose ? numLoose : num);
  }

  static clone(arr) {
    return dpArr(arr);
  }

  static indexes(arr) {
    return arr.map((_, i) => i);
  }

  static map(arr, fn, l) {
    l = l || arr.length;
    const vc = Array(l);

    for (--l; l >= 0b0; l--) vc[l] = fn(arr[l], l);

    return vc;
  }

  static mutateMap(arr, fn, l) {
    l = l || arr.length;

    for (--l; l >= 0b0; l--) arr[l] = fn(arr[l], l);

    return arr;
  }

  static select(arr, indexes, hi) {
    hi = hi || indexes.length;
    const vc = Array(hi);

    for (--hi; hi >= 0b0; hi--) vc[hi] = arr[indexes[hi]];

    return vc;
  }
  /**
   *
   * @param {*[]} arr
   * @param {number[]} indexes - number indexes of the positions to be spliced, should be in ascending order.
   * @param {number} [hi]
   * @returns {*[]}
   */


  static splices(arr, indexes, hi) {
    hi = hi || indexes.length;

    for (--hi; hi >= 0b0; hi--) arr.splice(indexes[hi], 1);

    return arr;
  }
  /**
   * Returns an array built from the elements of a given set of arrays.
   * Each element of the returned array is determined by elements from every one of the array-set with the same index.
   * The returned array has length of the first array in the array set.
   * @param {function} zipper The function {x(i)|x(i) ∈ array(i), i ∈ 0,1...n-1, n=arraySet.size} -> y(i), where y
   * is the returned array
   * @param {*[][]} arraySet The array-set to determine the returned array.
   * @returns {*[]|undefined} array
   */


  static multiZip(zipper, ...arraySet) {
    const firstArray = arraySet[0];

    if (!!firstArray) {
      const [len, cnt] = [firstArray.length, arraySet.length];
      const result = Array(len);

      for (let i = 0; i < len; i++) {
        const params = Array(cnt);

        for (let j = 0; j < cnt; j++) {
          params[j] = arraySet[j][i];
        }

        result[i] = zipper(params);
      }

      return result;
    } else {
      return undefined;
    }
  }
  /**
   *
   * @param {number} size
   * @param {*} initial
   * @param {function} progress
   * @returns {*[]}
   */


  static progression(size, initial, progress) {
    switch (size) {
      case 0:
        return [];

      case 1:
        return [initial];

      default:
        const arr = new Array(size);
        arr[0] = initial;

        for (let i = 1; i < size; i++) {
          arr[i] = progress(arr[i - 1]);
        }

        return arr;
    }
  }

  static range(min, max, fn) {
    const len = max - min + 1,
          arr = new Array(len);

    if (fn && typeof fn === 'function') {
      for (let i = 0; i < len; i++) arr[i] = fn(min + i);
    } else {
      for (let i = 0; i < len; i++) arr[i] = min + i;
    }

    return arr;
  }
  /**
   * Create an arithmetic progression
   * @param {number} size
   * @param {number|string} initial
   * @param {number} delta
   * @returns {*[]}
   */


  static arithmetic(size, initial, delta) {
    return Ar.progression(size, initial, previous => previous + delta);
  }
  /**
   * Create a geometric progression
   * @param {number} size
   * @param {number} initial
   * @param {number} ratio
   * @returns {*[]}}
   */


  static geometric(size, initial, ratio) {
    return Ar.progression(size, initial, previous => previous * ratio);
  }
  /**
   *
   * @param {*[]} ar1
   * @param {*[]} ar2
   * @param {function} product
   * @returns {*[]}
   */


  static decartes(ar1, ar2, product) {
    const l1 = ar1.length,
          l2 = ar2.length;
    let arr = Array(l1 * l2);

    for (let i = 0, j, k = 0; i < l1; i++) {
      for (j = 0; j < l2; j++) {
        arr[k++] = product(ar1[i], ar2[j]);
      }
    }

    return arr; // for (let x of ar1) {
    //   arr.push(...ar2.map(y => product(x, y)))
    // }
  }

  static randSample(arr) {
    // const len = arr.length
    // switch (len) {
    //   case 0:
    //     return undefined
    //   case 1:
    //     return arr[0]
    //   default:
    //     const idx = Roulett.rand(0, len)
    //     return arr[idx]
    // }
    return arr[~~(Math.random() * arr.length)];
  }

  static take(arr, len) {
    return arr.slice(0, len);
  }

  static zip(arL, arR, zipper, l) {
    l = l || arL.length;
    const vc = Array(l);

    for (--l; l >= 0; l--) vc[l] = zipper(arL[l], arR[l], l);

    return vc;
  }

} // Array.prototype.zip = function (another, zipper) {


class PalettTable {
  static meta() {
    return {
      colors: ColorGroups.entire,
      degrees: Degrees.entire,
      colorGroups: Object.keys(ColorGroups),
      degreeGroups: Object.keys(Degrees)
    };
  }
  /**
   *
   * @param {string} [space='hex]
   * @param {string[]} [degrees=Degrees.entire]
   * @param {string[]} [colors=ColorGroups.entire]
   * @param {boolean} [average=false]
   * @param {boolean} [cellColor=false]
   * @returns {CrosTab|{side:string,banner:string,matrix:*[][]}}
   */


  static crosTab({
    space = 'hex',
    degrees = Degrees.entire,
    colors = ColorGroups.entire,
    average = false,
    cellColor = false
  } = {}) {
    var _space, _space2, _Ob$selectValues$map;

    const formatter = (_space = space, toFormatter(_space, cellColor));
    const dataPicker = (_space2 = space, toDataPicker(_space2));
    const matrix = (_Ob$selectValues$map = veho.Ob.selectValues(palett.Palett, colors).map(tube => veho.Ob.selectValues(tube, degrees).map(dataPicker)), veho.Mx.transpose(_Ob$selectValues$map));
    const crosTab = new CrosTab(degrees, colors.map(readify), veho.Mx.map(matrix, formatter));

    if (average) {
      var _space3;

      const statistic = (_space3 = space, toStatistic(_space3));

      const formattedStatistic = ar => {
        var _ref, _ar;

        return _ref = (_ar = ar, statistic(_ar)), formatter(_ref);
      };

      const averageOnColumns = veho.Mx.columns(matrix, formattedStatistic);
      crosTab.unshiftRow('average', averageOnColumns);
      const averageOnRows = [''].concat(Ar.map(matrix, formattedStatistic));
      crosTab.unshiftCol('average', averageOnRows);
    }

    return crosTab;
  }

}

function _defineProperty$2(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

const xPalett = PalettTable.crosTab({
  degrees: Degrees.readable,
  colors: ColorGroups.rainbow
});

class PalettSelector {
  static random() {
    const color = Roulett.element(xPalett.banner);
    const degree = Roulett.element(xPalett.side);
    const hex = xPalett.queryCell(degree, color);
    return {
      color,
      degree,
      hex
    };
  }

}

_defineProperty$2(PalettSelector, "pool", xPalett.ht * xPalett.wd);

const Directs = {
  p_wise: 0,
  r_wise: 1,
  c_wise: 2
};

Object.defineProperty(exports, 'Hsl', {
  enumerable: true,
  get: function () {
    return farbe.Hsl;
  }
});
Object.defineProperty(exports, 'Rgb', {
  enumerable: true,
  get: function () {
    return farbe.Rgb;
  }
});
Object.defineProperty(exports, 'greys', {
  enumerable: true,
  get: function () {
    return palett.Greys;
  }
});
Object.defineProperty(exports, 'palette', {
  enumerable: true,
  get: function () {
    return palett.Palett;
  }
});
Object.defineProperty(exports, 'shades', {
  enumerable: true,
  get: function () {
    return palett.Shades;
  }
});
exports.ColorCards = PalettTable;
exports.Directs = Directs;
exports.Pal = Pal;
exports.Visual = Visual;
